#!/usr/local/cpanel/3rdparty/bin/perl

# bhs3-cpbackup.pl
# Copyright 2015 Beyond Hosting LLC
#
# Based on code from scripts/custom_backup_destination.pl.skeleton and Cpanel/Transport/Files/AmazonS3.pm
#                                                 Copyright(c) 2013 cPanel, Inc.
#                                                           All rights Reserved.
# copyright@cpanel.net                                         http://cpanel.net
# This code is subject to the cPanel license. Unauthorized copying is prohibited

use strict;
use warnings;

use File::Spec               ();
use Amazon::S3               ();

# These are the commands that a custom destination script must process
my %commands = (
    put    => \&bhs3_put,
    get    => \&bhs3_get,
    ls     => \&bhs3_ls,
    mkdir  => \&bhs3_mkdir,
    chdir  => \&bhs3_chdir,
    rmdir  => \&bhs3_rmdir,
    delete => \&bhs3_delete,
);

# There must be at least the command and the local directory
usage() if ( @ARGV < 2 );

#
# The command line arguments passed to the script will be in the following order:
# command, local_directory, command arguments, and optionally, host, user password
# The local directory is passed in so we know from which directory to run the command
# we need to pass this in each time since we start the script fresh for each command
#
my ( $cmd, $local_dir, @args ) = @ARGV;

# complain if the command does not exist
usage() unless exists $commands{$cmd};

# Run our command
$commands{$cmd}->(@args);

#
# This script should only really be executed by the custom backup destination type
# If someone executes it directly out of curiosity, give them usage info
#
sub usage {
    my @cmds = sort keys %commands;
    print STDERR "This script is for implementing a custom backup destination\n";
    print STDERR "It requires the following arguments:  cmd, local_dir, cmd_args\n";
    print STDERR "These are the valid commands:  @cmds\n";
    exit 1;
}

#
# Connect to S3 Gateway and return Amazon::S3 object
#

sub bhs3_connect {
    my ( $host, $access_key ) = @_;

    my $secret_key = $ENV{'PASSWORD'};

    my $bucket = "";

    # Custom destinations only allow the hostname to be specified, extracting bucket from the configured hostname
    if ($host =~ /^([a-z0-9][a-z0-9-]*)\.s3\.beyondhosting\.net$/i) {
        $bucket = $1;
        $host =~ s/^[a-z0-9][a-z0-9-]*\.//;
    } else {
        print STDERR "Could not detect bucket name from hostname. Please use format: <bucket>.s3.beyondhosting.net\n";
        exit 1;
    }

    my $s3 = Amazon::S3->new(
        {
            'host'                  => $host,
            'aws_access_key_id'     => $access_key,
            'aws_secret_access_key' => $secret_key,
            'retry'                 => 1,
            'secure'                => 1,
        }
    );

    # If we did not create it, die
    if ( !$s3 ) {
        print STDERR "Could not connect to the S3 Gateway\n";
        exit 1;
    }

    my $bucket_obj = $s3->bucket($bucket);

    return ($s3, $bucket_obj);
}

#
# These functions were taken from the AmazonS3 Transport Destination Module (AmazonS3.pm)
#

#
# Remove leading/trailing/double slashes from path
#
sub _sanitize_remote_path {
    my ($path) = @_;

    # Break up the path pieces & remove an blanks due to leading/trailing/double slashes
    my @chunks = grep { $_ } File::Spec->splitdir($path);

    # Put the path chunks back together - minus the detrius
    return File::Spec->catdir(@chunks);
}

#
# This portion contains the implementations for the various commands
# that the script needs to support in order to implement a custom destination
#

#
# Copy a local file to a remote destination
#
sub bhs3_put {
    my ( $local, $remote, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);


    # The size we break the file into for multipart upload
    # We hardcode this for now; but we may make this a config
    # option if there is demand for it.
    my $upload_chunk_size = 20 * 1024**2;

    $remote = _sanitize_remote_path($remote);

    # If a file is over a threshold, we use the multipart API to upload the file.
    # Where as Amazon has a hard requirement to use the multipart API for files
    # 5Gb or greater, we found the multipart api is less prone to error for
    # uploading large files in a lower bandwidth situation.
    # (In particular, sending files to a geographicaly distant S3 bucket.)
    # So, if the file size is larger than a single chunk used in a
    # multipart upload, then use the multipart upload
    my $file_size = ( -s $local || 0 );
    if ( $file_size > $upload_chunk_size ) {
        my $chunk_size = $upload_chunk_size;

        # Amazon only allows 10,000 chunks in a multipart upload
        # So, if the file is bigger than 10,000 times our chunk size
        # then we'll just increase the chunk size.
        my $max_size = 10_000 * $upload_chunk_size;
        if ( $file_size > $max_size ) {

            # Ensure that we'll never have more than 10,000 parts by making the
            # chunk slightly larger.
            $chunk_size = ( $file_size / 10_000 ) + 1;
        }

        my ( $res, $msg ) = bhs3_do_multipart_upload( $bucket, $local, $remote, $chunk_size );

        if ($res) {
            print "Successfully transported backup to S3\n";
        }
        else {
            print STDERR "Could not upload the object to the S3 bucket. Error: $msg\n";
            exit 1;
        }
    }
    elsif ( $bucket->add_key_filename( $remote, $local ) ) {

        # If less then 5Gb, use the normal put operation to upload it
        print "Successfully transported backup to S3\n";
    }
    else {
        print STDERR "S3 Error: " . ( $s3->err || '' ) . ": " . ( $s3->errstr || '' );
        print STDERR "Could not upload the object to the S3 bucket\n";
        exit 1;
    }

    return;
}

#
# Copy a remote file to a local destination
#
sub bhs3_get {
    my ( $remote, $local, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);

    $remote = _sanitize_remote_path($remote);

    if ( $bucket->get_key_filename( $remote, 'GET', $local ) ) {
        print "Successfully retrieved backup from S3: $local\n";
    }
    else {
        print STDERR "Error retrieving backup from S3\n";
        exit 1;
    }

    return;
}

#
# Print out the results of doing an ls operation
# The calling program will expect the data to be
# in the format supplied by 'ls -l' and have it
# printed to STDOUT
#
sub bhs3_ls {
    my ( $path, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);

    $path = _sanitize_remote_path($path) . '/';

    my $res = $bucket->list_all();

    my %contents;
    foreach my $item ( @{ $res->{'keys'} } ) {

        # Slip the ones that dont math our path
        next unless ( !$path or $path eq '/' or $item->{'key'} =~ m/^$path/ );

        my $type = ( $item->{'size'} == 0 and $item->{'key'} =~ m|/$| ) ? 'd' : '-';
        $item->{'key'} =~ s/^$path//;

        # This can happen if the key and the path are equal
        next unless $item->{'key'};

        $contents{ $item->{'key'} } = {
            'type'  => $type,
            'owner' => $item->{'owner_displayname'},
            'size'  => $item->{'size'},
        };
    }

    # In Amazon S3, a file can be in a folder without the folder existing on it's own
    # This is kind of strange, we need to return the existence of these folders
    # The pruning code needs to find the container folders for old backups to delete
    # So, we are going to go through each item and, if need be, create parent items
    my @file_list = keys %contents;    # Get the keys before starting the loop; we'll be modding the hash in the loop
    foreach my $file (@file_list) {

        my $file_info = $contents{$file};

        # Go up the tree of parents until reaching the top
        my $parent = $file;
        while ( $parent = File::Basename::dirname($parent) and $parent ne '.' and $parent ne '/' ) {

            # Make sure it has a trailing slash, S3's folders have this
            $parent .= '/' unless $parent =~ m|/$|;

            # Skip if it is already there
            next if $contents{$parent};

            # Add an entry for the non-existent parent
            # The parent, will, of course, be a directory
            $contents{$parent} = {
                'type'  => 'd',
                'owner' => $file_info->{'owner'},
                'size'  => '0',
            };
        }
    }

    # Again, since Amazon S3 list gives all the items in the bucket,
    # Our list of items contains some stuff nested several layers deep.
    # We only want to return the immediate contents of the specified folder
    # We could filter this out until now since we needed to do the operation
    # above to get the parent folders of the nested items
    my @nested_items = grep { /\/.+$/ } keys %contents;
    delete @contents{@nested_items};

    # The output must look like the results of "ls -l" & contain perms for Historical Reasons
    my @ls = map { "$contents{$_}{'type'}rw-r--r-- X $contents{$_}{'owner'} X $contents{$_}{'size'} X X X $_" } sort keys %contents;

    print join("\n", @ls);

    return;
}

#
# Create a directory on the remote destination
#
sub bhs3_mkdir {
    my ( $path, $recurse, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);

    # Folders in Amazon S3 end with a slash
    $path = _sanitize_remote_path($path) . '/';

    # A folder, in Amazon S3, is a key ending in a '/' with no contents
    if ( $bucket->add_key( $path, '' ) ) {
        print "Successfully created S3 folder\n";
    }
    else {
        print STDERR "Error creating directory in S3\n";
        exit 1;
    }

    return;
}

#
# Change into a directory on the remote destination
# This does not have the same meaning as it normally would since the script
# is run anew for each command call.
# This needs to do the operation to ensure it doesn't fail
# then print the new resulting directory that the calling program
# will pass in as the local directory for subsequent calls
#
sub bhs3_chdir {
    my ( $path, $host, $user ) = @_;
    # There is nothing to do for S3
    return;
}

#
# Recursively delete a directory on the remote destination
#
sub bhs3_rmdir {
    my ( $path, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);

    # Folders in Amazon S3 end with a slash
    $path = _sanitize_remote_path($path) . '/';

    # Amazon S3 doesn't list contents of folders, just the whole bucket
    my $res = $bucket->list_all();

    # Find all the keys "under" this one
    # Do a reverse sort so we delete the objects in the folder
    # before deleting the folder
    my @keys = reverse sort grep { $_ =~ m/^$path/ } map { $_->{'key'} } @{ $res->{'keys'} };

    foreach my $key (@keys) {

        if (!$bucket->delete_key($key)) {
            print STDERR "Error: Could not delete $key from S3\n";
            exit 1;
        }
    }

    print "Successfully deleted $path\n";

    return;
}

#
# Delete an individual file on the remote destination
#
sub bhs3_delete {
    my ( $path, $host, $user ) = @_;

    my ($s3, $bucket) = bhs3_connect($host, $user);

    $path = _sanitize_remote_path($path);

    if (!$bucket->delete_key($path)) {
        print STDERR "Error: Could not delete $path from S3\n";
        exit 1;
    }

    print "Successfully deleted $path from S3\n";

    return;
}

#
# Amazon requires that objects larger than 5Gb be uploaded
# in chunks via the multipart upload API
#
sub bhs3_do_multipart_upload {
    my ( $bucket, $local, $remote, $chunk_size ) = @_;

    # Start multipart transfer
    my $upload_id;
    eval { $upload_id = $bucket->initiate_multipart_upload($remote); };

    # initiate will die if it fails; but, we also want to verify
    # that we got an actual return value as well (in case AWS's response was malformed)
    return ( 0, "Failure in initiate_multipart_upload: $@" ) if $@;
    return ( 0, "Unable to initiate multipart upload" ) unless $upload_id;

    # Now handle the file upload
    my $fh;
    eval {
        open( $fh, '<', $local ) or die "Could not open $local: $!";
        binmode($fh) or die "Could not set $local to binary mode:  $!";

        my $part_num = 1;
        my ( $buffer, $length, %parts_hash );

        while ( ( $length = read( $fh, $buffer, $chunk_size ) ) != 0 ) {

            # Upload the chunk we just read
            my $etag = $bucket->upload_part_of_multipart_upload( $remote, $upload_id, $part_num, $buffer, $length );

            # We should get an etag back as an acknowlegement of the upload
            die "Error uploading part $part_num of $local" unless $etag;

            # Save the etag, we'll need to send a hash of part_numbers => etags
            # when we want to complete the upload process
            $parts_hash{ $part_num++ } = $etag;
        }

        # read will return undef if there is an error (vs zero for end of file)
        die "Failure reading $local:  $!" unless defined $length;

        close($fh);

        # Notify Amazon of the completion of the file upload
        $bucket->complete_multipart_upload( $remote, $upload_id, \%parts_hash )
          or die "Unable to finalize the file upload";
    };
    if ($@) {
        close($fh) if $fh;
        my $error_message = $@;

        # Abort the upload operation if a failure occurred somewhere
        eval { $bucket->abort_multipart_upload( $remote, $upload_id ); };
        if ($@) {
            $error_message .= '; And error aborting upload:  ' . $@;
        }

        return ( 0, $error_message );
    }

    return ( 1, 'OK' );
}